"""PyCayenneLPP: A decoder and encoder for the CayenneLPP data format.

PyCayenneLPP offers a concise interface with proper encoding and decoding
functionality for the CayenneLPP format, supporting many sensor types.
"""

from .lpp_data import LppData
from .lpp_frame import LppFrame
from .lpp_util import LppUtil

__all__ = ['LppData', 'LppFrame', 'LppUtil']



from .lpp_data import LppData


class LppFrame(object):
    """A LPP frame instance.

    A LppFrame can hold multiple LppData objects to be encoded or
    decoded at once. It also provides convenient helper functions
    to easily add sensor values of certain type.

    Attributes:
        data (list): a list of LppData objects
        maxsize (int): (optional) byte size limit
    """

    def __init__(self, data=None, maxsize=0):
        """Create a LppFrame object with (optional) arguments."""
        self._maxsize = maxsize
        self._data = []
        if data:
            for d in data:
                self.__add_data_item(d)

    def __str__(self):
        """Return a pretty string representation of the LppFrame object."""
        out = "LppFrame(data = ["
        if self._data:
            out = out + "\n"
        for d in self._data:
            out = out + "  " + str(d) + "\n"
        out = out + "])"
        return out

    def __len__(self):
        """Return the number of LppData items in this LppFrame."""
        return len(self._data)

    def __iter__(self):
        """Return an iterator over all LppData items in this LppFrame."""
        count = 0
        while count < len(self._data):
            yield self._data[count]
            count += 1

    def __bytes__(self):
        """Return this LppFrame object as a byte string."""
        buf = bytearray()
        for d in self._data:
            buf += bytes(d)
        return bytes(buf)

    def to_bytes(self):
        """Return this LppFrame object as a byte string (for MicroPython)."""
        buf = bytearray()
        for d in self._data:
            buf += d.to_bytes()
        return bytes(buf)

    @classmethod
    def from_bytes(cls, buf):
        """Parse a given byte string and return as a LppFrame object."""
        i = 0
        data = []
        while i < len(buf):
            lppdata = LppData.from_bytes(buf[i:])
            data.append(lppdata)
            i = i + len(lppdata)
        return cls(data)

    def __add_data_item(self, item):
        """Helper function to add an LppData item to this LppFrame."""
        if not isinstance(item, LppData):
            raise TypeError()
        if self.maxsize > 0:
            if self.size + len(item) > self.maxsize:
                raise BufferError()
        self._data.append(item)

    @property
    def data(self):
        """Return list of data items."""
        return self._data

    @property
    def maxsize(self):
        """Return max allowed byte size for this LppFrame."""
        return self._maxsize

    @maxsize.setter
    def maxsize(self, value):
        """Set the max byte size for this LppFrame."""
        if value < 0:
            raise ValueError("Maxsize must be positive integer.")
        if value > 0 and value < self.size:
            raise ValueError("Maxsize must be greater than current frame size")
        self._maxsize = value

    def reset(self):
        """Reset LppFrame by clearing the list of LppData items."""
        self._data.clear()

    @property
    def size(self):
        """Return the length of the LppFrame byte string representation."""
        size = 0
        for d in self._data:
            size += len(d)
        return size

    def get_by_type(self, type_):
        """Return sub list of data with items matching given type."""
        return [d for d in self.data if int(d.type) == type_]

    def get_by_name(self, name):
        """Return sub list of data with items matching given name."""
        name = name.strip().lower()
        return [d for d in self.data if str(d.type).lower().startswith(name)]

    def add_by_type(self, type_, channel, value_tuple):
        """Generic helper to add LppDate to this LppFrame."""
        if not isinstance(value_tuple, tuple):
            raise TypeError('Parameter (value_tuple) must be a tuple!')
        data = LppData(channel, type_, value_tuple)
        self.__add_data_item(data)

    def add_digital_input(self, channel, value):
        """Create and add a digital input LppData item."""
        self.add_by_type(0, channel, (value, ))

    def add_digital_output(self, channel, value):
        """Create and add a digital output LppData item."""
        self.add_by_type(1, channel, (value, ))

    def add_analog_input(self, channel, value):
        """Create and add an analog input LppData item."""
        self.add_by_type(2, channel, (value, ))

    def add_analog_output(self, channel, value):
        """Create and add an analog output LppData item."""
        self.add_by_type(3, channel, (value, ))

    def add_generic(self, channel, value):
        """Create and add a generic 4-byte unsigned integer LppData item."""
        self.add_by_type(100, channel, (value, ))

    def add_luminosity(self, channel, value):
        """Create and add an illuminance sensor LppData item."""
        self.add_by_type(101, channel, (value, ))

    def add_presence(self, channel, value):
        """Create and add a presence sensor LppData item."""
        self.add_by_type(102, channel, (value, ))

    def add_temperature(self, channel, value):
        """Create and add a temperature sensor LppData item."""
        self.add_by_type(103, channel, (value, ))

    def add_humidity(self, channel, value):
        """Create and add a humidity sensor LppData item."""
        self.add_by_type(104, channel, (value, ))

    def add_accelerometer(self, channel, x, y, z):
        """Create and add a accelerometer sensor LppData item."""
        self.add_by_type(113, channel, (x, y, z))

    def add_pressure(self, channel, value):
        """Alias method for add_barometer()."""
        self.add_barometer(channel, value)

    def add_barometer(self, channel, value):
        """Create and add a barometer sensor LppData item."""
        self.add_by_type(115, channel, (value, ))

    def add_voltage(self, channel, value):
        """Create and add a voltage sensor LppData item."""
        self.add_by_type(116, channel, (value, ))

    def add_current(self, channel, value):
        """Create and add a current sensor LppData item."""
        self.add_by_type(117, channel, (value, ))

    def add_frequency(self, channel, value):
        """Create and add a frequency sensor LppData item."""
        self.add_by_type(118, channel, (value, ))

    def add_percentage(self, channel, value):
        """Create and add a percentage LppData item."""
        self.add_by_type(120, channel, (value, ))

    def add_altitude(self, channel, value):
        """Create and add a altitude LppData item."""
        self.add_by_type(121, channel, (value, ))

    def add_load(self, channel, value):
        """Create and add a load sensor LppData item."""
        self.add_by_type(122, channel, (value, ))

    def add_concentration(self, channel, value):
        """Create and add a concentration LppData item."""
        self.add_by_type(125, channel, (value, ))

    def add_power(self, channel, value):
        """Create and add a power sensor LppData item."""
        self.add_by_type(128, channel, (value, ))

    def add_distance(self, channel, value):
        """Create and add a distance LppData item."""
        self.add_by_type(130, channel, (value, ))

    def add_energy(self, channel, value):
        """Create and add a energy sensor LppData item."""
        self.add_by_type(131, channel, (value, ))

    def add_direction(self, channel, value):
        """Create and add a direction LppData item."""
        self.add_by_type(132, channel, (value, ))

    def add_unix_time(self, channel, value):
        """Create and add a unix timestamp LppData item."""
        self.add_by_type(133, channel, (value, ))

    def add_gyrometer(self, channel, x, y, z):
        """Create and add a gyrometer sensor LppData item."""
        self.add_by_type(134, channel, (x, y, z))

    def add_colour(self, channel, red, green, blue):
        """Create and add a color sensor LppData item."""
        self.add_by_type(135, channel, (red, green, blue))

    def add_gps(self, channel, lat, lon, alt):
        """Alias method for add_location()."""
        self.add_location(channel, lat, lon, alt)

    def add_location(self, channel, lat, lon, alt):
        """Create and add a location LppData item."""
        self.add_by_type(136, channel, (lat, lon, alt))

    def add_switch(self, channel, value):
        """Create and add a switch LppData item."""
        self.add_by_type(142, channel, (value, ))



class LppType(object):
    """Cayenne LPP type object.

    The LppType provides a simple wrapper to all sensor types of
    the Cayenne LPP standard (and beyond). It ensures proper
    encoding and decoding with sensible checks. This class is for
    internal use only and thus is not (directly) exposed.

    Attributes:
        type (int): LPP type ID number
        name (str): human readable type description
        sizes (list): byte size of values
        scales (list): scaling of values
        signs (list): signess of values
    """

    __lpp_types = {
        0:      ('Digital Input', [1], [1], [False]),
        1:      ('Digital Output', [1], [1], [False]),
        2:      ('Analog Input', [2], [100], [True]),
        3:      ('Analog Output', [2], [100], [True]),
        100:    ('Generic Sensor', [4], [1], [False]),
        101:    ('Illuminance', [2], [1], [False]),
        102:    ('Presence', [1], [1], [False]),
        103:    ('Temperature', [2], [10], [True]),
        104:    ('Humidity', [1], [2], [False]),
        113:    ('Accelerometer', [2, 2, 2], [1000, 1000, 1000],
                 [True, True, True]),
        115:    ('Barometer', [2], [10], [False]),
        116:    ('Voltage', [2], [100], [False]),
        117:    ('Current', [2], [1000], [False]),
        118:    ('Frequency', [4], [1], [False]),
        120:    ('Percentage', [1], [1], [False]),
        121:    ('Altitude', [2], [1], [True]),
        122:    ('Load', [3], [1000], [True]),
        125:    ('Concentration', [2], [1], [False]),
        128:    ('Power', [2], [1], [False]),
        130:    ('Distance', [4], [1000], [False]),
        131:    ('Energy', [4], [1000], [False]),
        132:    ('Direction', [2], [1], [False]),
        133:    ('Time', [4], [1], [False]),
        134:    ('Gyrometer', [2, 2, 2], [100, 100, 100],
                 [True, True, True]),
        135:    ('Colour', [1, 1, 1], [1, 1, 1],
                 [False, False, False]),
        136:    ('Location', [3, 3, 3], [10000, 10000, 100],
                 [True, True, True]),
        142:    ('Switch', [1], [1], [False])
    }

    def __init__(self, type_, name, sizes, scales, signs):
        """Create a LppType object with given attributes."""
        if not isinstance(type_, int):
            raise TypeError('Parameter (type_) must be an integer!')
        if not isinstance(name, str):
            raise TypeError('Parameter (name) must be a string!')
        if not isinstance(sizes, list):
            raise TypeError('Parameter (sizes) must be a list of integers!')
        if not isinstance(scales, list):
            raise TypeError('Parameter (scales) must be a list of integers!')
        if not isinstance(signs, list):
            raise TypeError('Parameter (signs) must be a list of integers!')
        if len(sizes) != len(scales) or len(scales) != len(signs):
            raise ValueError('Invalid parameter length: sizes, scales, signs!')
        self.type = type_
        self.name = name
        self.sizes = sizes
        self.scales = scales
        self.signs = signs

    def __int__(self):
        """Return LppType as integer, i.e. its numeric type."""
        return self.type

    def __str__(self):
        """Return LppType as string, i.e. its name."""
        return self.name

    @staticmethod
    def __assert_data_tuple(data, num):
        """Internal helper to ensure data is a tuple of given `num` length."""
        if not isinstance(data, tuple):
            data = (data,)
        if not len(data) == num:
            raise ValueError()
        return data

    @staticmethod
    def __from_bytes(buf):
        """Internal helper to parse a number from buffer."""
        buflen = len(buf)
        val = 0
        for i in range(buflen):
            shift = (buflen - i - 1) * 8
            val |= buf[i] << shift
        return val

    @staticmethod
    def __to_bytes(val, buflen):
        """Internal helper to write a value to a buffer."""
        buf = bytearray(buflen)
        val = int(val)
        for i in range(buflen):
            shift = (buflen - i - 1) * 8
            buf[i] = (val >> shift) & 0xff
        return buf

    @staticmethod
    def __to_signed(val, size):
        """Internal helper to convert unsigned int to signed."""
        mask = 0x00
        for i in range(size):
            mask |= 0xff << (i * 8)
        if val >= (1 << ((size * 8) - 1)):
            val = -1 - (val ^ mask)
        return val

    @staticmethod
    def __to_unsigned(val):
        """Convert signed (two's complement) value to unsigned."""
        if val < 0:
            val = ~(-val - 1)
        return val

    @classmethod
    def get_lpp_type(cls, type_):
        """Return LppType object for given type or `None` if not found."""
        if not isinstance(type_, int):
            raise TypeError('Parameter (type_) must be an integer!')

        if type_ in cls.__lpp_types:
            return cls(type_, *cls.__lpp_types[type_])
        return None

    @property
    def dimension(self):
        """Return number of value dimensions."""
        return len(self.sizes)

    @property
    def size(self):
        """Return size of byte string representation."""
        return sum(self.sizes)

    def decode(self, buf):
        """Parse LppType from a byte string."""
        if len(buf) != sum(self.sizes):
            raise BufferError('Invalid buffer length!')
        data = []
        pos = 0
        for i in range(len(self.sizes)):
            size = self.sizes[i]
            value = self.__from_bytes(buf[pos:(pos + size)])
            if self.signs[i]:
                value = self.__to_signed(value, size)
            value = value / self.scales[i]
            data.append(value)
            pos += size
        return tuple(data)

    def encode(self, data):
        """Convert LppType into a byte string."""
        data = self.__assert_data_tuple(data, len(self.sizes))
        buf = bytearray(sum(self.sizes))
        pos = 0
        for i in range(len(self.sizes)):
            size = self.sizes[i]
            value = data[i]
            if not self.signs[i] and value < 0:
                raise ValueError('Invalid data, must be non negative!')
            value = int(value * self.scales[i])
            maxval = 1 << (size * 8)
            if value >= maxval:
                raise ValueError('Invalid data, exceed value range!')
            if self.signs[i]:
                value = self.__to_unsigned(value)
            buf[pos:(pos + size)] = self.__to_bytes(value, size)
            pos += size
        return buf



from .lpp_data import LppData
from .lpp_frame import LppFrame
from .lpp_type import LppType


class LppUtil():
    """Cayenne LPP utility functions.

    This class provides helper functions and wrapper, e.g. for encoding and
    decoding LppType, LppData, and LppFrame into various data formats.
    """

    @staticmethod
    def json_encode(obj, type2str=False):
        """Encode LppType, LppData, and LppFrame to JSON."""
        if isinstance(obj, LppType):
            if type2str:
                return obj.name
            return obj.type
        if isinstance(obj, LppData):
            return obj.__dict__
        if isinstance(obj, LppFrame):
            return obj.data
        raise TypeError(repr(obj) + " is not JSON serialized")

    @classmethod
    def json_encode_type_int(cls, obj):
        """Wrapper function encode type as int in JSON."""
        return cls.json_encode(obj, False)

    @classmethod
    def json_encode_type_str(cls, obj):
        """Wrapper function encode type as string in JSON."""
        return cls.json_encode(obj, True)



from setuptools import setup


def readme():
    with open('README.md') as f:
        return f.read()


setup(
    name='pycayennelpp',
    version='2.4.0',
    python_requires='>=3.6',
    description='Encoder and Decoder for CayenneLLP',
    long_description=readme(),
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9'
    ],
    keywords='cayenne lpp iot lora lorawan ttn',
    url='http://github.com/smlng/pycayennelpp',
    author='smlng',
    author_email='s@mlng.net',
    license='MIT',
    packages=['cayennelpp'],
    setup_requires=["pytest-runner"],
    tests_require=['pytest'],
    include_package_data=True
)
